#
#   BLIS
#   An object-based framework for developing high-performance BLAS-like
#   libraries.
#
#   Copyright (C) 2018, The University of Texas at Austin
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are
#   met:
#    - Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    - Neither the name(s) of the copyright holder(s) nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#   HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# CPU: Intel Xeon X5550, 2666 MHz
#
00000000 ******** => 0000000B 756E6547 6C65746E 49656E69
00000001 ******** => 000106A2 00100800 00BCE3BD BFEBFBFF
00000002 ******** => 55035A01 00F0B2E4 00000000 09CA212C
00000003 ******** => 00000000 00000000 00000000 00000000
00000005 ******** => 00000040 00000040 00000003 00021120
00000006 ******** => 00000003 00000002 00000001 00000000
00000007 ******** => 00000000 00000000 00000000 00000000
00000008 ******** => 00000000 00000000 00000000 00000000
00000009 ******** => 00000000 00000000 00000000 00000000
0000000A ******** => 07300403 00000000 00000000 00000603
80000000 ******** => 80000008 00000000 00000000 00000000
80000001 ******** => 00000000 00000000 00000001 28100000
80000002 ******** => 756E6547 20656E69 65746E49 2952286C
80000003 ******** => 55504320 20202020 20202020 40202020
80000004 ******** => 30303020 20402030 37362E32 007A4847
80000005 ******** => 00000000 00000000 00000000 00000000
80000006 ******** => 00000000 00000000 01006040 00000000
80000007 ******** => 00000000 00000000 00000000 00000100
80000008 ******** => 00003028 00000000 00000000 00000000
