EXPORTS
bli_abort
bli_absqsc
bli_acquire_mij
bli_acquire_mpart
bli_acquire_mpart_b2t
bli_acquire_mpart_br2tl
bli_acquire_mpart_l2r
bli_acquire_mpart_mdim
bli_acquire_mpart_mndim
bli_acquire_mpart_ndim
bli_acquire_mpart_r2l
bli_acquire_mpart_t2b
bli_acquire_mpart_tl2br
bli_acquire_vi
bli_acquire_vpart_b2f
bli_acquire_vpart_f2b
bli_addd
bli_addd_ex
bli_addm
bli_addm_ex
bli_addsc
bli_addv
bli_addv_ex
bli_align_dim_to_mult
bli_align_dim_to_size
bli_align_ptr_to_size
bli_amaxv
bli_amaxv_ex
bli_arch_query_id
bli_arch_string
bli_asumv
bli_asumv_ex
bli_axpbyv
bli_axpbyv_ex
bli_axpy2v
bli_axpy2v_ex
bli_axpyd
bli_axpyd_ex
bli_axpyf
bli_axpyf_ex
bli_axpym
bli_axpym_ex
bli_axpyv
bli_axpyv_ex
bli_blksz_create
bli_blksz_create_ed
bli_blksz_free
bli_blksz_init
bli_blksz_init_easy
bli_blksz_init_ed
bli_cabsqsc
bli_caddd
bli_caddd_ex
bli_caddm
bli_caddm_ex
bli_caddsc
bli_caddv
bli_caddv_ex
bli_camaxv
bli_camaxv_ex
bli_castm
bli_castnzm
bli_castv
bli_casumv
bli_casumv_ex
bli_caxpbyv
bli_caxpbyv_ex
bli_caxpy2v
bli_caxpy2v_ex
bli_caxpyd
bli_caxpyd_ex
bli_caxpyf
bli_caxpyf_ex
bli_caxpym
bli_caxpym_ex
bli_caxpyv
bli_caxpyv_ex
bli_cccastm
bli_cccastnzm
bli_cccastv
bli_cccopysc
bli_ccopyd
bli_ccopyd_ex
bli_ccopym
bli_ccopym_ex
bli_ccopyv
bli_ccopyv_ex
bli_ccxpbym_md
bli_ccxpbym_md_ex
bli_cdcastm
bli_cdcastnzm
bli_cdcastv
bli_cdcopysc
bli_cdivsc
bli_cdotaxpyv
bli_cdotaxpyv_ex
bli_cdotv
bli_cdotv_ex
bli_cdotxaxpyf
bli_cdotxaxpyf_ex
bli_cdotxf
bli_cdotxf_ex
bli_cdotxv
bli_cdotxv_ex
bli_cdxpbym_md
bli_cdxpbym_md_ex
bli_ceqm
bli_ceqsc
bli_ceqv
bli_cfprintm
bli_cfprintv
bli_cgemm
bli_cgemm_ex
bli_cgemmt
bli_cgemmt_ex
bli_cgemv
bli_cgemv_ex
bli_cger
bli_cger_ex
bli_cgetijm
bli_cgetijv
bli_cgetsc
bli_cgesc
bli_cgtsc
bli_check_error_code_helper
bli_chemm
bli_chemm_ex
bli_chemv
bli_chemv_ex
bli_cher
bli_cher2
bli_cher2_ex
bli_cher2k
bli_cher2k_ex
bli_cher_ex
bli_cherk
bli_cherk_ex
bli_cinvertd
bli_cinvertd_ex
bli_cinvertsc
bli_cinvertv
bli_cinvertv_ex
bli_cinvscald
bli_cinvscald_ex
bli_cinvscalm
bli_cinvscalm_ex
bli_cinvscalv
bli_cinvscalv_ex
bli_clock
bli_clock_min_diff
bli_clesc
bli_cltsc
bli_cmachval
bli_cmkherm
bli_cmkherm_ex
bli_cmksymm
bli_cmksymm_ex
bli_cmktrim
bli_cmktrim_ex
bli_cmulsc
bli_cnorm1m
bli_cnorm1m_ex
bli_cnorm1v
bli_cnorm1v_ex
bli_cnormfm
bli_cnormfm_ex
bli_cnormfsc
bli_cnormfv
bli_cnormfv_ex
bli_cnormim
bli_cnormim_ex
bli_cnormiv
bli_cnormiv_ex
bli_cntl_clear_node
bli_cntl_copy
bli_cntl_create_node
bli_cntl_free
bli_cntl_free_node
bli_cntl_mark_family
bli_cntx_clear
bli_cntx_print
bli_cntx_set_blkszs
bli_cntx_set_ind_blkszs
bli_cntx_set_l3_sup_handlers
bli_cntx_set_ukr_prefs
bli_cntx_set_ukrs
bli_copyd
bli_copyd_ex
bli_copym
bli_copym_ex
bli_copysc
bli_copyv
bli_copyv_ex
bli_cprintm
bli_cprintv
bli_crandm
bli_crandm_ex
bli_crandnm
bli_crandnm_ex
bli_crandnv
bli_crandnv_ex
bli_crandv
bli_crandv_ex
bli_cscal2d
bli_cscal2d_ex
bli_cscal2m
bli_cscal2m_ex
bli_cscal2v
bli_cscal2v_ex
bli_cscald
bli_cscald_ex
bli_cscalm
bli_cscalm_ex
bli_cscalv
bli_cscalv_ex
bli_cscastm
bli_cscastnzm
bli_cscastv
bli_cscopysc
bli_csetd
bli_csetd_ex
bli_csetid
bli_csetid_ex
bli_csetijm
bli_csetijv
bli_csetm
bli_csetm_ex
bli_csetsc
bli_csetv
bli_csetv_ex
bli_cshiftd
bli_cshiftd_ex
bli_csqrtrsc
bli_csqrtsc
bli_csubd
bli_csubd_ex
bli_csubm
bli_csubm_ex
bli_csubsc
bli_csubv
bli_csubv_ex
bli_csumsqv
bli_csumsqv_ex
bli_cswapv
bli_cswapv_ex
bli_csxpbym_md
bli_csxpbym_md_ex
bli_csymm
bli_csymm_ex
bli_csymv
bli_csymv_ex
bli_csyr
bli_csyr2
bli_csyr2_ex
bli_csyr2k
bli_csyr2k_ex
bli_csyr_ex
bli_csyrk
bli_csyrk_ex
bli_ctrmm
bli_ctrmm3
bli_ctrmm3_ex
bli_ctrmm_ex
bli_ctrmv
bli_ctrmv_ex
bli_ctrsm
bli_ctrsm_ex
bli_ctrsv
bli_ctrsv_ex
bli_cunzipsc
bli_cxpbyd
bli_cxpbyd_ex
bli_cxpbym
bli_cxpbym_ex
bli_cxpbyv
bli_cxpbyv_ex
bli_czcastm
bli_czcastnzm
bli_czcastv
bli_czcopysc
bli_czipsc
bli_czxpbym_md
bli_czxpbym_md_ex
bli_dabsqsc
bli_daddd
bli_daddd_ex
bli_daddm
bli_daddm_ex
bli_daddsc
bli_daddv
bli_daddv_ex
bli_damaxv
bli_damaxv_ex
bli_dasumv
bli_dasumv_ex
bli_daxpbyv
bli_daxpbyv_ex
bli_daxpy2v
bli_daxpy2v_ex
bli_daxpyd
bli_daxpyd_ex
bli_daxpyf
bli_daxpyf_ex
bli_daxpym
bli_daxpym_ex
bli_daxpyv
bli_daxpyv_ex
bli_dccastm
bli_dccastnzm
bli_dccastv
bli_dccopysc
bli_dcopyd
bli_dcopyd_ex
bli_dcopym
bli_dcopym_ex
bli_dcopyv
bli_dcopyv_ex
bli_dcxpbym_md
bli_dcxpbym_md_ex
bli_ddcastm
bli_ddcastnzm
bli_ddcastv
bli_ddcopysc
bli_ddivsc
bli_ddotaxpyv
bli_ddotaxpyv_ex
bli_ddotv
bli_ddotv_ex
bli_ddotxaxpyf
bli_ddotxaxpyf_ex
bli_ddotxf
bli_ddotxf_ex
bli_ddotxv
bli_ddotxv_ex
bli_ddxpbym_md
bli_ddxpbym_md_ex
bli_deqm
bli_deqsc
bli_deqv
bli_dfprintm
bli_dfprintv
bli_dgemm
bli_dgemm_ex
bli_dgemmt
bli_dgemmt_ex
bli_dgemv
bli_dgemv_ex
bli_dger
bli_dger_ex
bli_dgetijm
bli_dgetijv
bli_dgetsc
bli_dgesc
bli_dgtsc
bli_dhemm
bli_dhemm_ex
bli_dhemv
bli_dhemv_ex
bli_dher
bli_dher2
bli_dher2_ex
bli_dher2k
bli_dher2k_ex
bli_dher_ex
bli_dherk
bli_dherk_ex
bli_dinvertd
bli_dinvertd_ex
bli_dinvertsc
bli_dinvertv
bli_dinvertv_ex
bli_dinvscald
bli_dinvscald_ex
bli_dinvscalm
bli_dinvscalm_ex
bli_dinvscalv
bli_dinvscalv_ex
bli_divsc
bli_dlesc
bli_dltsc
bli_dmachval
bli_dmkherm
bli_dmkherm_ex
bli_dmksymm
bli_dmksymm_ex
bli_dmktrim
bli_dmktrim_ex
bli_dmulsc
bli_dnorm1m
bli_dnorm1m_ex
bli_dnorm1v
bli_dnorm1v_ex
bli_dnormfm
bli_dnormfm_ex
bli_dnormfsc
bli_dnormfv
bli_dnormfv_ex
bli_dnormim
bli_dnormim_ex
bli_dnormiv
bli_dnormiv_ex
bli_dotaxpyv
bli_dotaxpyv_ex
bli_dotv
bli_dotv_ex
bli_dotxaxpyf
bli_dotxaxpyf_ex
bli_dotxf
bli_dotxf_ex
bli_dotxv
bli_dotxv_ex
bli_dprintm
bli_dprintv
bli_drandm
bli_drandm_ex
bli_drandnm
bli_drandnm_ex
bli_drandnv
bli_drandnv_ex
bli_drandv
bli_drandv_ex
bli_dscal2d
bli_dscal2d_ex
bli_dscal2m
bli_dscal2m_ex
bli_dscal2v
bli_dscal2v_ex
bli_dscald
bli_dscald_ex
bli_dscalm
bli_dscalm_ex
bli_dscalv
bli_dscalv_ex
bli_dscastm
bli_dscastnzm
bli_dscastv
bli_dscopysc
bli_dsetd
bli_dsetd_ex
bli_dsetid
bli_dsetid_ex
bli_dsetijm
bli_dsetijv
bli_dsetm
bli_dsetm_ex
bli_dsetsc
bli_dsetv
bli_dsetv_ex
bli_dshiftd
bli_dshiftd_ex
bli_dsqrtrsc
bli_dsqrtsc
bli_dsubd
bli_dsubd_ex
bli_dsubm
bli_dsubm_ex
bli_dsubsc
bli_dsubv
bli_dsubv_ex
bli_dsumsqv
bli_dsumsqv_ex
bli_dswapv
bli_dswapv_ex
bli_dsxpbym_md
bli_dsxpbym_md_ex
bli_dsymm
bli_dsymm_ex
bli_dsymv
bli_dsymv_ex
bli_dsyr
bli_dsyr2
bli_dsyr2_ex
bli_dsyr2k
bli_dsyr2k_ex
bli_dsyr_ex
bli_dsyrk
bli_dsyrk_ex
bli_dt_size
bli_dt_string
bli_dtrmm
bli_dtrmm3
bli_dtrmm3_ex
bli_dtrmm_ex
bli_dtrmv
bli_dtrmv_ex
bli_dtrsm
bli_dtrsm_ex
bli_dtrsv
bli_dtrsv_ex
bli_dunzipsc
bli_dxpbyd
bli_dxpbyd_ex
bli_dxpbym
bli_dxpbym_ex
bli_dxpbyv
bli_dxpbyv_ex
bli_dzcastm
bli_dzcastnzm
bli_dzcastv
bli_dzcopysc
bli_dzipsc
bli_dzxpbym_md
bli_dzxpbym_md_ex
bli_eqm
bli_eqsc
bli_eqv
bli_error_checking_level
bli_error_checking_level_set
bli_finalize
bli_fprintm
bli_fprintv
bli_free_user
bli_gemm
bli_gemm_ex
bli_gemm_ukernel
bli_gemmt
bli_gemmt_ex
bli_gemmtrsm_ukernel
bli_gemv
bli_gemv_ex
bli_ger
bli_ger_ex
bli_getijm
bli_getijv
bli_getopt
bli_getopt_init_state
bli_getsc
bli_gks_init_ref_cntx
bli_gks_l3_ukr_impl_string
bli_gks_l3_ukr_impl_type
bli_gks_query_cntx
bli_gks_query_ind_cntx
bli_gks_query_nat_cntx
bli_gesc
bli_gtsc
bli_hemm
bli_hemm_ex
bli_hemv
bli_hemv_ex
bli_her
bli_her2
bli_her2_ex
bli_her2k
bli_her2k_ex
bli_her_ex
bli_herk
bli_herk_ex
bli_ifprintm
bli_ifprintv
bli_igetsc
bli_ind_disable
bli_ind_disable_all
bli_ind_disable_all_dt
bli_ind_disable_dt
bli_ind_enable
bli_ind_enable_dt
bli_ind_oper_enable_only
bli_ind_oper_find_avail
bli_ind_oper_get_avail_impl_string
bli_info_get_blas_int_type_size
bli_info_get_enable_blas
bli_info_get_enable_cblas
bli_info_get_enable_hpx
bli_info_get_enable_hpx_as_default
bli_info_get_enable_memkind
bli_info_get_enable_openmp
bli_info_get_enable_openmp_as_default
bli_info_get_enable_pba_pools
bli_info_get_enable_pthreads
bli_info_get_enable_pthreads_as_default
bli_info_get_enable_sandbox
bli_info_get_enable_sba_pools
bli_info_get_enable_stay_auto_init
bli_info_get_enable_threading
bli_info_get_enable_tls
bli_info_get_gemm_impl_string
bli_info_get_gemm_ukr_impl_string
bli_info_get_gemmt_impl_string
bli_info_get_gemmtrsm_l_ukr_impl_string
bli_info_get_gemmtrsm_u_ukr_impl_string
bli_info_get_heap_addr_align_size
bli_info_get_heap_stride_align_size
bli_info_get_hemm_impl_string
bli_info_get_her2k_impl_string
bli_info_get_herk_impl_string
bli_info_get_int_type_size
bli_info_get_int_type_size_str
bli_info_get_max_type_size
bli_info_get_num_fp_types
bli_info_get_page_size
bli_info_get_pool_addr_align_size_a
bli_info_get_pool_addr_align_size_b
bli_info_get_pool_addr_align_size_c
bli_info_get_pool_addr_align_size_gen
bli_info_get_pool_addr_offset_size_a
bli_info_get_pool_addr_offset_size_b
bli_info_get_pool_addr_offset_size_c
bli_info_get_pool_addr_offset_size_gen
bli_info_get_simd_align_size
bli_info_get_simd_num_registers
bli_info_get_simd_size
bli_info_get_stack_buf_align_size
bli_info_get_stack_buf_max_size
bli_info_get_symm_impl_string
bli_info_get_syr2k_impl_string
bli_info_get_syrk_impl_string
bli_info_get_thread_jrir_rr
bli_info_get_thread_jrir_slab
bli_info_get_thread_jrir_tlb
bli_info_get_trmm3_impl_string
bli_info_get_trmm_impl_string
bli_info_get_trsm_impl_string
bli_info_get_trsm_l_ukr_impl_string
bli_info_get_trsm_u_ukr_impl_string
bli_info_get_version_str
bli_init
bli_invertd
bli_invertd_ex
bli_invertsc
bli_invertv
bli_invertv_ex
bli_invscald
bli_invscald_ex
bli_invscalm
bli_invscalm_ex
bli_invscalv
bli_invscalv_ex
bli_iprintm
bli_iprintv
bli_isetsc
bli_l3_thrinfo_create
bli_lesc
bli_ltsc
bli_machval
bli_malloc_user
bli_mkherm
bli_mkherm_ex
bli_mksymm
bli_mksymm_ex
bli_mktrim
bli_mktrim_ex
bli_mulsc
bli_norm1m
bli_norm1m_ex
bli_norm1v
bli_norm1v_ex
bli_normfm
bli_normfm_ex
bli_normfsc
bli_normfv
bli_normfv_ex
bli_normim
bli_normim_ex
bli_normiv
bli_normiv_ex
bli_obj_alloc_buffer
bli_obj_attach_buffer
bli_obj_create
bli_obj_create_1x1
bli_obj_create_1x1_with_attached_buffer
bli_obj_create_conf_to
bli_obj_create_with_attached_buffer
bli_obj_create_without_buffer
bli_obj_free
bli_obj_print
bli_obj_scalar_apply_scalar
bli_obj_scalar_attach
bli_obj_scalar_cast_to
bli_obj_scalar_detach
bli_obj_scalar_init_detached
bli_obj_scalar_init_detached_copy_of
bli_obj_scalar_reset
bli_pack_get_pack_a
bli_pack_get_pack_b
bli_pack_set_pack_a
bli_pack_set_pack_b
bli_packm_alloc
bli_packm_alloc_ex
bli_packm_blk_var1
bli_packm_scalar
bli_param_map_blis_to_char_conj
bli_param_map_blis_to_char_diag
bli_param_map_blis_to_char_dt
bli_param_map_blis_to_char_side
bli_param_map_blis_to_char_trans
bli_param_map_blis_to_char_uplo
bli_param_map_blis_to_netlib_diag
bli_param_map_blis_to_netlib_machval
bli_param_map_blis_to_netlib_side
bli_param_map_blis_to_netlib_trans
bli_param_map_blis_to_netlib_uplo
bli_param_map_char_to_blis_conj
bli_param_map_char_to_blis_diag
bli_param_map_char_to_blis_dt
bli_param_map_char_to_blis_side
bli_param_map_char_to_blis_trans
bli_param_map_char_to_blis_uplo
bli_pba_query
bli_printm
bli_printv
bli_projm
bli_projv
bli_pthread_barrier_destroy
bli_pthread_barrier_init
bli_pthread_barrier_wait
bli_pthread_cond_broadcast
bli_pthread_cond_destroy
bli_pthread_cond_init
bli_pthread_cond_wait
bli_pthread_create
bli_pthread_join
bli_pthread_mutex_destroy
bli_pthread_mutex_init
bli_pthread_mutex_lock
bli_pthread_mutex_trylock
bli_pthread_mutex_unlock
bli_pthread_once
bli_randm
bli_randm_ex
bli_randnm
bli_randnm_ex
bli_randnv
bli_randnv_ex
bli_randv
bli_randv_ex
bli_rntm_init_from_global
bli_rntm_set_num_threads
bli_rntm_set_ways
bli_rntm_set_ways_for_op
bli_sabsqsc
bli_saddd
bli_saddd_ex
bli_saddm
bli_saddm_ex
bli_saddsc
bli_saddv
bli_saddv_ex
bli_samaxv
bli_samaxv_ex
bli_sasumv
bli_sasumv_ex
bli_saxpbyv
bli_saxpbyv_ex
bli_saxpy2v
bli_saxpy2v_ex
bli_saxpyd
bli_saxpyd_ex
bli_saxpyf
bli_saxpyf_ex
bli_saxpym
bli_saxpym_ex
bli_saxpyv
bli_saxpyv_ex
bli_scal2d
bli_scal2d_ex
bli_scal2m
bli_scal2m_ex
bli_scal2v
bli_scal2v_ex
bli_scald
bli_scald_ex
bli_scalm
bli_scalm_ex
bli_scalv
bli_scalv_ex
bli_sccastm
bli_sccastnzm
bli_sccastv
bli_sccopysc
bli_scopyd
bli_scopyd_ex
bli_scopym
bli_scopym_ex
bli_scopyv
bli_scopyv_ex
bli_scxpbym_md
bli_scxpbym_md_ex
bli_sdcastm
bli_sdcastnzm
bli_sdcastv
bli_sdcopysc
bli_sdivsc
bli_sdotaxpyv
bli_sdotaxpyv_ex
bli_sdotv
bli_sdotv_ex
bli_sdotxaxpyf
bli_sdotxaxpyf_ex
bli_sdotxf
bli_sdotxf_ex
bli_sdotxv
bli_sdotxv_ex
bli_sdxpbym_md
bli_sdxpbym_md_ex
bli_seqm
bli_seqsc
bli_seqv
bli_setd
bli_setd_ex
bli_setid
bli_setid_ex
bli_setijm
bli_setijv
bli_setim
bli_setiv
bli_setm
bli_setm_ex
bli_setrm
bli_setrv
bli_setsc
bli_setv
bli_setv_ex
bli_sfprintm
bli_sfprintv
bli_sgemm
bli_sgemm_ex
bli_sgemmt
bli_sgemmt_ex
bli_sgemv
bli_sgemv_ex
bli_sger
bli_sger_ex
bli_sgetijm
bli_sgetijv
bli_sgetsc
bli_sgesc
bli_sgtsc
bli_shemm
bli_shemm_ex
bli_shemv
bli_shemv_ex
bli_sher
bli_sher2
bli_sher2_ex
bli_sher2k
bli_sher2k_ex
bli_sher_ex
bli_sherk
bli_sherk_ex
bli_shiftd
bli_shiftd_ex
bli_sinvertd
bli_sinvertd_ex
bli_sinvertsc
bli_sinvertv
bli_sinvertv_ex
bli_sinvscald
bli_sinvscald_ex
bli_sinvscalm
bli_sinvscalm_ex
bli_sinvscalv
bli_sinvscalv_ex
bli_sleep
bli_slesc
bli_sltsc
bli_smachval
bli_smkherm
bli_smkherm_ex
bli_smksymm
bli_smksymm_ex
bli_smktrim
bli_smktrim_ex
bli_smulsc
bli_snorm1m
bli_snorm1m_ex
bli_snorm1v
bli_snorm1v_ex
bli_snormfm
bli_snormfm_ex
bli_snormfsc
bli_snormfv
bli_snormfv_ex
bli_snormim
bli_snormim_ex
bli_snormiv
bli_snormiv_ex
bli_sprintm
bli_sprintv
bli_sqrtrsc
bli_sqrtsc
bli_srandm
bli_srandm_ex
bli_srandnm
bli_srandnm_ex
bli_srandnv
bli_srandnv_ex
bli_srandv
bli_srandv_ex
bli_sscal2d
bli_sscal2d_ex
bli_sscal2m
bli_sscal2m_ex
bli_sscal2v
bli_sscal2v_ex
bli_sscald
bli_sscald_ex
bli_sscalm
bli_sscalm_ex
bli_sscalv
bli_sscalv_ex
bli_sscastm
bli_sscastnzm
bli_sscastv
bli_sscopysc
bli_ssetd
bli_ssetd_ex
bli_ssetid
bli_ssetid_ex
bli_ssetijm
bli_ssetijv
bli_ssetm
bli_ssetm_ex
bli_ssetsc
bli_ssetv
bli_ssetv_ex
bli_sshiftd
bli_sshiftd_ex
bli_ssqrtrsc
bli_ssqrtsc
bli_ssubd
bli_ssubd_ex
bli_ssubm
bli_ssubm_ex
bli_ssubsc
bli_ssubv
bli_ssubv_ex
bli_ssumsqv
bli_ssumsqv_ex
bli_sswapv
bli_sswapv_ex
bli_ssxpbym_md
bli_ssxpbym_md_ex
bli_ssymm
bli_ssymm_ex
bli_ssymv
bli_ssymv_ex
bli_ssyr
bli_ssyr2
bli_ssyr2_ex
bli_ssyr2k
bli_ssyr2k_ex
bli_ssyr_ex
bli_ssyrk
bli_ssyrk_ex
bli_strmm
bli_strmm3
bli_strmm3_ex
bli_strmm_ex
bli_strmv
bli_strmv_ex
bli_strsm
bli_strsm_ex
bli_strsv
bli_strsv_ex
bli_subd
bli_subd_ex
bli_subm
bli_subm_ex
bli_subsc
bli_subv
bli_subv_ex
bli_sumsqv
bli_sumsqv_ex
bli_sunzipsc
bli_swapv
bli_swapv_ex
bli_sxpbyd
bli_sxpbyd_ex
bli_sxpbym
bli_sxpbym_ex
bli_sxpbyv
bli_sxpbyv_ex
bli_symm
bli_symm_ex
bli_symv
bli_symv_ex
bli_syr
bli_syr2
bli_syr2_ex
bli_syr2k
bli_syr2k_ex
bli_syr_ex
bli_syrk
bli_syrk_ex
bli_szcastm
bli_szcastnzm
bli_szcastv
bli_szcopysc
bli_szipsc
bli_szxpbym_md
bli_szxpbym_md_ex
bli_thrcomm_barrier
bli_thrcomm_bcast
bli_thread_get_ic_nt
bli_thread_get_ir_nt
bli_thread_get_jc_nt
bli_thread_get_jr_nt
bli_thread_get_num_threads
bli_thread_get_pc_nt
bli_thread_get_thread_impl
bli_thread_get_thread_impl_str
bli_thread_launch
bli_thread_range_sub
bli_thread_set_num_threads
bli_thread_set_num_threads_
bli_thread_set_thread_impl
bli_thread_set_ways
bli_thread_set_ways_
bli_thrinfo_free
bli_trmm
bli_trmm3
bli_trmm3_ex
bli_trmm_ex
bli_trmv
bli_trmv_ex
bli_trsm
bli_trsm_ex
bli_trsm_ukernel
bli_trsv
bli_trsv_ex
bli_unzipsc
bli_xpbyd
bli_xpbyd_ex
bli_xpbym
bli_xpbym_ex
bli_xpbym_md
bli_xpbym_md_ex
bli_xpbyv
bli_xpbyv_ex
bli_zabsqsc
bli_zaddd
bli_zaddd_ex
bli_zaddm
bli_zaddm_ex
bli_zaddsc
bli_zaddv
bli_zaddv_ex
bli_zamaxv
bli_zamaxv_ex
bli_zasumv
bli_zasumv_ex
bli_zaxpbyv
bli_zaxpbyv_ex
bli_zaxpy2v
bli_zaxpy2v_ex
bli_zaxpyd
bli_zaxpyd_ex
bli_zaxpyf
bli_zaxpyf_ex
bli_zaxpym
bli_zaxpym_ex
bli_zaxpyv
bli_zaxpyv_ex
bli_zccastm
bli_zccastnzm
bli_zccastv
bli_zccopysc
bli_zcopyd
bli_zcopyd_ex
bli_zcopym
bli_zcopym_ex
bli_zcopyv
bli_zcopyv_ex
bli_zcxpbym_md
bli_zcxpbym_md_ex
bli_zdcastm
bli_zdcastnzm
bli_zdcastv
bli_zdcopysc
bli_zdivsc
bli_zdotaxpyv
bli_zdotaxpyv_ex
bli_zdotv
bli_zdotv_ex
bli_zdotxaxpyf
bli_zdotxaxpyf_ex
bli_zdotxf
bli_zdotxf_ex
bli_zdotxv
bli_zdotxv_ex
bli_zdxpbym_md
bli_zdxpbym_md_ex
bli_zeqm
bli_zeqsc
bli_zeqv
bli_zfprintm
bli_zfprintv
bli_zgemm
bli_zgemm_ex
bli_zgemmt
bli_zgemmt_ex
bli_zgemv
bli_zgemv_ex
bli_zger
bli_zger_ex
bli_zgetijm
bli_zgetijv
bli_zgetsc
bli_zgesc
bli_zgtsc
bli_zhemm
bli_zhemm_ex
bli_zhemv
bli_zhemv_ex
bli_zher
bli_zher2
bli_zher2_ex
bli_zher2k
bli_zher2k_ex
bli_zher_ex
bli_zherk
bli_zherk_ex
bli_zinvertd
bli_zinvertd_ex
bli_zinvertsc
bli_zinvertv
bli_zinvertv_ex
bli_zinvscald
bli_zinvscald_ex
bli_zinvscalm
bli_zinvscalm_ex
bli_zinvscalv
bli_zinvscalv_ex
bli_zipsc
bli_zlesc
bli_zltsc
bli_zmachval
bli_zmkherm
bli_zmkherm_ex
bli_zmksymm
bli_zmksymm_ex
bli_zmktrim
bli_zmktrim_ex
bli_zmulsc
bli_znorm1m
bli_znorm1m_ex
bli_znorm1v
bli_znorm1v_ex
bli_znormfm
bli_znormfm_ex
bli_znormfsc
bli_znormfv
bli_znormfv_ex
bli_znormim
bli_znormim_ex
bli_znormiv
bli_znormiv_ex
bli_zprintm
bli_zprintv
bli_zrandm
bli_zrandm_ex
bli_zrandnm
bli_zrandnm_ex
bli_zrandnv
bli_zrandnv_ex
bli_zrandv
bli_zrandv_ex
bli_zscal2d
bli_zscal2d_ex
bli_zscal2m
bli_zscal2m_ex
bli_zscal2v
bli_zscal2v_ex
bli_zscald
bli_zscald_ex
bli_zscalm
bli_zscalm_ex
bli_zscalv
bli_zscalv_ex
bli_zscastm
bli_zscastnzm
bli_zscastv
bli_zscopysc
bli_zsetd
bli_zsetd_ex
bli_zsetid
bli_zsetid_ex
bli_zsetijm
bli_zsetijv
bli_zsetm
bli_zsetm_ex
bli_zsetsc
bli_zsetv
bli_zsetv_ex
bli_zshiftd
bli_zshiftd_ex
bli_zsqrtrsc
bli_zsqrtsc
bli_zsubd
bli_zsubd_ex
bli_zsubm
bli_zsubm_ex
bli_zsubsc
bli_zsubv
bli_zsubv_ex
bli_zsumsqv
bli_zsumsqv_ex
bli_zswapv
bli_zswapv_ex
bli_zsxpbym_md
bli_zsxpbym_md_ex
bli_zsymm
bli_zsymm_ex
bli_zsymv
bli_zsymv_ex
bli_zsyr
bli_zsyr2
bli_zsyr2_ex
bli_zsyr2k
bli_zsyr2k_ex
bli_zsyr_ex
bli_zsyrk
bli_zsyrk_ex
bli_ztrmm
bli_ztrmm3
bli_ztrmm3_ex
bli_ztrmm_ex
bli_ztrmv
bli_ztrmv_ex
bli_ztrsm
bli_ztrsm_ex
bli_ztrsv
bli_ztrsv_ex
bli_zunzipsc
bli_zxpbyd
bli_zxpbyd_ex
bli_zxpbym
bli_zxpbym_ex
bli_zxpbyv
bli_zxpbyv_ex
bli_zzcastm
bli_zzcastnzm
bli_zzcastv
bli_zzcopysc
bli_zzipsc
bli_zzxpbym_md
bli_zzxpbym_md_ex
caxpby_
caxpy_
cblas_caxpby
cblas_caxpy
cblas_ccopy
cblas_cdotc_sub
cblas_cdotu_sub
cblas_cgbmv
cblas_cgemm
cblas_cgemm3m
cblas_cgemm_batch
cblas_cgemmt
cblas_cgemmtr
cblas_cgemv
cblas_cgerc
cblas_cgeru
cblas_chbmv
cblas_chemm
cblas_chemv
cblas_cher
cblas_cher2
cblas_cher2k
cblas_cherk
cblas_chpmv
cblas_chpr
cblas_chpr2
cblas_cscal
cblas_csscal
cblas_cswap
cblas_csymm
cblas_csyr2k
cblas_csyrk
cblas_ctbmv
cblas_ctbsv
cblas_ctpmv
cblas_ctpsv
cblas_ctrmm
cblas_ctrmv
cblas_ctrsm
cblas_ctrsv
cblas_dasum
cblas_daxpby
cblas_daxpy
cblas_dcopy
cblas_ddot
cblas_dgbmv
cblas_dgemm
cblas_dgemm_batch
cblas_dgemmt
cblas_dgemmtr
cblas_dgemv
cblas_dger
cblas_dnrm2
cblas_drot
cblas_drotg
cblas_drotm
cblas_drotmg
cblas_dsbmv
cblas_dscal
cblas_dsdot
cblas_dspmv
cblas_dspr
cblas_dspr2
cblas_dswap
cblas_dsymm
cblas_dsymv
cblas_dsyr
cblas_dsyr2
cblas_dsyr2k
cblas_dsyrk
cblas_dtbmv
cblas_dtbsv
cblas_dtpmv
cblas_dtpsv
cblas_dtrmm
cblas_dtrmv
cblas_dtrsm
cblas_dtrsv
cblas_dzasum
cblas_dznrm2
cblas_icamax
cblas_idamax
cblas_isamax
cblas_izamax
cblas_sasum
cblas_saxpby
cblas_saxpy
cblas_scasum
cblas_scnrm2
cblas_scopy
cblas_sdot
cblas_sdsdot
cblas_sgbmv
cblas_sgemm
cblas_sgemm_batch
cblas_sgemmt
cblas_sgemmtr
cblas_sgemv
cblas_sger
cblas_snrm2
cblas_srot
cblas_srotg
cblas_srotm
cblas_srotmg
cblas_ssbmv
cblas_sscal
cblas_sspmv
cblas_sspr
cblas_sspr2
cblas_sswap
cblas_ssymm
cblas_ssymv
cblas_ssyr
cblas_ssyr2
cblas_ssyr2k
cblas_ssyrk
cblas_stbmv
cblas_stbsv
cblas_stpmv
cblas_stpsv
cblas_strmm
cblas_strmv
cblas_strsm
cblas_strsv
cblas_xerbla
cblas_zaxpby
cblas_zaxpy
cblas_zcopy
cblas_zdotc_sub
cblas_zdotu_sub
cblas_zdscal
cblas_zgbmv
cblas_zgemm
cblas_zgemm3m
cblas_zgemm_batch
cblas_zgemmt
cblas_zgemmtr
cblas_zgemv
cblas_zgerc
cblas_zgeru
cblas_zhbmv
cblas_zhemm
cblas_zhemv
cblas_zher
cblas_zher2
cblas_zher2k
cblas_zherk
cblas_zhpmv
cblas_zhpr
cblas_zhpr2
cblas_zscal
cblas_zswap
cblas_zsymm
cblas_zsyr2k
cblas_zsyrk
cblas_ztbmv
cblas_ztbsv
cblas_ztpmv
cblas_ztpsv
cblas_ztrmm
cblas_ztrmv
cblas_ztrsm
cblas_ztrsv
ccopy_
cdotc_
cdotcsub_
cdotu_
cdotusub_
cgbmv_
cgemm3m_
cgemm_
cgemm_batch_
cgemmt_
cgemmtr_
cgemv_
cgerc_
cgeru_
chbmv_
chemm_
chemv_
cher2_
cher2k_
cher_
cherk_
chpmv_
chpr2_
chpr_
crotg_
cscal_
csrot_
csscal_
cswap_
csymm_
csyr2k_
csyrk_
ctbmv_
ctbsv_
ctpmv_
ctpsv_
ctrmm_
ctrmv_
ctrsm_
ctrsv_
dasum_
dasumsub_
daxpby_
daxpy_
dcabs1_
dcopy_
ddot_
ddotsub_
dgbmv_
dgemm_
dgemm_batch_
dgemmt_
dgemmtr_
dgemv_
dger_
dnrm2_
dnrm2sub_
drot_
drotg_
drotm_
drotmg_
dsbmv_
dscal_
dsdot_
dsdotsub_
dspmv_
dspr2_
dspr_
dswap_
dsymm_
dsymv_
dsyr2_
dsyr2k_
dsyr_
dsyrk_
dtbmv_
dtbsv_
dtpmv_
dtpsv_
dtrmm_
dtrmv_
dtrsm_
dtrsv_
dzasum_
dzasumsub_
dznrm2_
dznrm2sub_
icamax_
icamaxsub_
idamax_
idamaxsub_
isamax_
isamaxsub_
izamax_
izamaxsub_
lsame_
sasum_
sasumsub_
saxpby_
saxpy_
scabs1_
scasum_
scasumsub_
scnrm2_
scnrm2sub_
scopy_
sdot_
sdotsub_
sdsdot_
sdsdotsub_
sgbmv_
sgemm_
sgemm_batch_
sgemmt_
sgemmtr_
sgemv_
sger_
snrm2_
snrm2sub_
srot_
srotg_
srotm_
srotmg_
ssbmv_
sscal_
sspmv_
sspr2_
sspr_
sswap_
ssymm_
ssymv_
ssyr2_
ssyr2k_
ssyr_
ssyrk_
stbmv_
stbsv_
stpmv_
stpsv_
strmm_
strmv_
strsm_
strsv_
xerbla_
xerbla_array_
zaxpby_
zaxpy_
zcopy_
zdotc_
zdotcsub_
zdotu_
zdotusub_
zdrot_
zdscal_
zgbmv_
zgemm3m_
zgemm_
zgemm_batch_
zgemmt_
zgemmtr_
zgemv_
zgerc_
zgeru_
zhbmv_
zhemm_
zhemv_
zher2_
zher2k_
zher_
zherk_
zhpmv_
zhpr2_
zhpr_
zrotg_
zscal_
zswap_
zsymm_
zsyr2k_
zsyrk_
ztbmv_
ztbsv_
ztpmv_
ztpsv_
ztrmm_
ztrmv_
ztrsm_
ztrsv_
