
#ifndef COMMON_H
#define COMMON_H

// enumerate all datatypes that will be tested
enum DATATYPES {
    DOUBLE  ,
    SINGLE  ,
    FLOAT16 ,
    BFLOAT16,
    INT16   ,
    INT8    ,
    INT4       
};

#endif