













% skx

plot_dom_all(2.0,32,1,'../../skx/st',      '../../skx/st_noxmem',     '../../skx/st_merged',     'skx_t1');close all;clear all
plot_dom_all(2.0,32,26,'../../skx/jc2ic13','../../skx/jc2ic13_noxmem','../../skx/jc2ic13_merged','skx_t26');close all;clear all
plot_dom_all(2.0,32,52,'../../skx/jc4ic13','../../skx/jc4ic13_noxmem','../../skx/jc4ic13_merged','skx_t52');close all;clear all

plot_dt_select(2.0,32,1,'../../skx/st',      '../../skx/st_noxmem',     '../../skx/st_merged',     'skx_t1');close all;clear all
plot_dt_select(2.0,32,26,'../../skx/jc2ic13','../../skx/jc2ic13_noxmem','../../skx/jc2ic13_merged','skx_t26');close all;clear all
plot_dt_select(2.0,32,52,'../../skx/jc4ic13','../../skx/jc4ic13_noxmem','../../skx/jc4ic13_merged','skx_t52');close all;clear all

% tx2

plot_dom_all(2.2,8,1,'../../tx2/st',      '../../tx2/st_noxmem',     '../../tx2/st_merged',     'tx2_t1');close all;clear all
plot_dom_all(2.2,8,28,'../../tx2/jc4ic7','../../tx2/jc4ic7_noxmem','../../tx2/jc4ic7_merged','tx2_t28');close all;clear all
plot_dom_all(2.2,8,56,'../../tx2/jc8ic7','../../tx2/jc8ic7_noxmem','../../tx2/jc8ic7_merged','tx2_t56');close all;clear all

plot_dt_select(2.2,8,1,'../../tx2/st',      '../../tx2/st_noxmem',     '../../tx2/st_merged',  'tx2_t1');close all;clear all
plot_dt_select(2.2,8,28,'../../tx2/jc4ic7','../../tx2/jc4ic7_noxmem','../../tx2/jc4ic7_merged','tx2_t28');close all;clear all
plot_dt_select(2.2,8,56,'../../tx2/jc8ic7','../../tx2/jc8ic7_noxmem','../../tx2/jc8ic7_merged','tx2_t56');close all;clear all
